package minijava.ast;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import minijava.TabSimb;

public class Classe {
	public String nome;
	public String pai;
	public List<Var> campos;
	public List<Metodo> metodos;
	public int lin;
	
	// Tipos de todos os campos, incluindo campos nas superclasses
	public Map<String, String> todosCampos;
	// Todos os métodos, incluindo métodos nas superclasses
	public Map<String, Metodo> todosMetodos;
	
	public Classe(String _nome, List<Var> _campos, List<Metodo> _metodos, int _lin) {
		nome = _nome;
		pai = "Object";
		campos = _campos;
		metodos = _metodos;
		lin = _lin;
	}

	public Classe(String _nome, String _pai, List<Var> _campos, List<Metodo> _metodos, int _lin) {
		nome = _nome;
		pai = _pai;
		campos = _campos;
		metodos = _metodos;
		lin = _lin;
	}

	public String toString() {
		String res = "class " + nome + " extends " + pai + "{\n";
		for(Var campo: campos) {
			res += campo + ";\n";
		}
		for(Metodo metodo: metodos) {
			res += metodo + "\n";
		}
		res += "}";
		return res;
	}
		
	public void adicionaCampos() {
		// TODO: implemente esse método
		// Adiciona os campos dessa classe e todas suas superclasses e
		// seus respectivos tipos no mapa "todosCampos"
		// Ter dois campos com o mesmo nome é um erro; caso uma subclasse
		// declare um campo com o mesmo nome de um campo de uma de suas
		// superclasses, o erro deve ser acusado na linha que está
		// redeclarando o campo na subclasse
	}
	
	public void adicionaMetodos() {
		if(todosMetodos != null) return;
		Classe cpai = Prog.classes.get(pai);
		if(cpai != null) {
			cpai.adicionaMetodos();
			todosMetodos = new HashMap<String, Metodo>(cpai.todosMetodos);
		} else todosMetodos = new HashMap<String, Metodo>();
		for(Metodo metodo: metodos) {
			if(todosMetodos.containsKey(metodo.nome))
				throw new RuntimeException("método " + metodo.nome + " redeclarado na linha " + metodo.lin);
			Metodo mpai = todosMetodos.get(metodo.nome);
			if(mpai != null) {
				if(mpai.params.size() != metodo.params.size())
					throw new RuntimeException("método " + metodo.nome + 
							" redefinido com assinatura diferente na linha " + metodo.lin);
				for(int i = 0; i < mpai.params.size(); i++) {
					if(!mpai.params.get(i).tipo.equals(metodo.params.get(i).tipo)) {
						throw new RuntimeException("método " + metodo.nome + 
								" redefinido com assinatura diferente na linha " + metodo.lin);
					}
				}
				if(!metodo.tret.equals(mpai.tret)) {
					throw new RuntimeException("método " + metodo.nome + 
							" redefinido com assinatura diferente na linha " + metodo.lin);
				}
			}
			todosMetodos.put(metodo.nome, metodo);
		}
	}
	
	public void tipos() {
		for(Metodo metodo: metodos) {
			TabSimb<String> escMetodo = new TabSimb<String>();
			escMetodo.inserir("this", this.nome);
			for(String campo: todosCampos.keySet()) {
				escMetodo.inserir(campo,  todosCampos.get(campo));
			}
			for(Var param: metodo.params) {
				if(!escMetodo.inserir(param.nome, param.tipo))
					throw new RuntimeException("parâmetro " + param.nome + " redeclarado na linha " + param.lin);
			}
			for(Var var: metodo.vars) {
				if(!escMetodo.inserir(var.nome, var.tipo))
					throw new RuntimeException("variável " + var.nome + 
							" redeclarada na linha " + var.lin);
			}
			for(Cmd cmd: metodo.cmds) {
				cmd.tipos(escMetodo);
			}
			String tret = metodo.ret.tipo(escMetodo);
			Tipo.subtipo(tret, metodo.tret, metodo.lin);
		}
	}

}
